;
;---------------------------------------------------------------
;
; Montezuma Micro CP/M v2.32 boot loader
; for Radio Shack Model 4 computer.
;
; The Model 4 boot rom loads the first sector (256 bytes)
; into memory at 4300h and executes it.
;
; This source code is for the Z80ASM assembler
; from SLR systems.
;
;---------------------------------------------------------------
;
	include	system.inc	; get system definitons
;
nmivec		equ	66h	; address of nmi vector
nmiadrs		equ	67h	; pointer to nmi interrupt routine
;
; Model 4 hardware ports
;
optreg		equ	84h
intmask		equ	0e0h
nmireg		equ	0e4h
modlatch	equ	0ech
fdcsel		equ	0f4h	; drive select port
;
; Floppy disk controller ports
;
fdccsr		equ	0f0h	; control and status
fdcsec		equ	0f2h	; sector register
fdcdat		equ	0f3h	; data register
;
; FDC commands (WD179x floppy disk controller)
;
readcmd		equ	80h
stepin		equ	54h
;
jumpcode	equ	0c3h
videobase	equ	0f800h
;
	org	4300h
;
boot:	ld	bc,ccpvec
	di
	ld	a,82h
	out	(optreg),a
	ld	hl,videobase
clrvid:	ld	(hl),20h	; clear video screen
	inc	hl
	ld	a,h
	or	l
	jr	nz,clrvid
	ld	sp,hl		; set stack to high memory
	out	(intmask),a	; disable nmi
	ld	a,40h
	out	(modlatch),a
	ld	a,3
	out	(optreg),a
	ld	a,jumpcode	; set up NMI vector
	ld	hl,nmiisr
	ld	(nmivec),a
	ld	(nmiadrs),hl
	ld	hl,(secsize)	; get sector (block) size
	ld	a,(seccount)	; get number of sectors to read
	ld	b,a
	ex	de,hl		; block size into DE
	ld	hl,(ccpvec)	; start address for CP/M system
	dec	h		; allow 256 bytes for boot sector
rdloop:	exx
	call	readsec		; read CP/M system into memory
	exx
	jp	nz,booterr	; go if read error
	add	hl,de		; else update buffer pointer
	djnz	rdloop
	ld	hl,(biosvec)	; done reading, get bios address
	jp	(hl)		; and execute bios code
;
; Read next sector
;
readsec:
	ld	hl,(sectblptr)
	ld	a,(hl)		; get next sector
	inc	a		; test for end of table
	jr	nz,rdsec2	; go if valid sector
	ld	hl,sectable	; else start at beginning of sector table
	ld	a,(side)	; swap disk sides
	ld	c,a
	ld	a,(sidemask)
	xor	c
	ld	(side),a
	jr	nz,rdsec2	; go if starting on side 1
	ld	a,stepin	; else step drive in one track
	call	issuecmd
rdsec1:	ld	a,81h		; select drive 0
	out	(fdcsel),a
	in	a,(fdccsr)
	and	1
	jr	nz,rdsec1	; wait for drive not busy
;
; read in next sector
;
rdsec2:	ld	a,(hl)		; get current sector number
	inc	hl
	ld	(sectblptr),hl	; update next sector to be read
	out	(fdcsec),a
	call	rdsec3		; read the sector
	ret	z		; done if no error
	call	rdsec3		; else try again
	ret	z		; done if no error, else try a third time
rdsec3:	ld	c,fdcdat	; init data port for ini
	ld	de,0c102h	; set up select data & drq mask
	exx			; restore buffer pointer in HL
	push	hl
	exx
	pop	hl
	ld	a,(side)	; add disk side to select data
	rlca
	rlca
	rlca
	rlca
	or	d
	ld	d,a
	and	0bfh
	out	(fdcsel),a	; select the drive and side
	ld	a,readcmd
	call	issuecmd	; start the read
	ld	a,0c0h		; enable nmi
	out	(nmireg),a
rdsec4:	in	a,(fdccsr)	; wait for drq from fdc
	and	e
	jr	z,rdsec4
	ini			; read first byte
	ld	a,d
rdsec5:	out	(fdcsel),a	; reselect drive
	ini			; and read data until nmi occurs
	jr	rdsec5		; (should be 512 bytes)
;
; NMI has occurred
;
nmiisr:	pop	hl		; take return address off the stack
	xor	a
	out	(nmireg),a	; disable nmi
	in	a,(fdccsr)	; get fdc status
	and	9ch		; mask off bits we don't need
	ret			; return to caller of readsec
;
; Send command to fdc chip and wait for busy bit to set
;
issuecmd:
	out	(fdccsr),a
bsywait:
	in	a,(fdccsr)
	and	1
	jr	z,bsywait
	ret
;
; Output boot error message to screen
; and loop forever
;
booterr:
	ld	a,2
	out	(optreg),a
	ld	hl,booterrmsg
	ld	de,videobase
	ld	bc,11
	ldir
hang:	jr	hang
;
ccpvec:		dw	ccpbase		; address to load system
biosvec:	dw	biosbase	; address to jump to when system loaded
;
secsize:	dw	200h		; 512 bytes per block
sidemask:	db	1
seccount:	db	20		; number of sectors to read
;
; Sector order on boot tracks
;
sectable:	db	1,3,5,7,9,2,4,6,8,0ah
		db	0ffh		; end of table marker
		db	0ffh,0ffh,0ffh,0ffh,0ffh,0ffh,0ffh,0ffh	; filler
		db	0ffh,0ffh,0ffh,0ffh,0ffh,0ffh,0ffh,0ffh
		db	0ffh,0ffh,0ffh
;
sectblptr:	dw	sectable
;
side:		db	0
;
booterrmsg:	db	'BOOT error!',0,0,0,0
;
	end
;

