
/*
 * CP/M file transfer utility.
 * Receives/transmits files from/to CP/M system at 9600 baud.
 * 01/25/04 - Modified for Linux
 * 05/22/04 - Added handling of X11 backspace (0x7f)
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>
#include	<dirent.h>
#include <sys/stat.h>
#include	<termios.h>
#include	<sys/ioctl.h>
#include <unistd.h>

#include	"serial.h"

#define	VERSION		1
#define	REVISION		1

#define	FALSE			0
#define	TRUE			1

#define	byte	unsigned char
#define	word	unsigned short

#define	ESC			0x1b
#define	ACK			0x06
#define	NAK			0x15
#define	CTRLZ			0x1a
#define	BACKSP		8
#define	X11BACKSP	0x7f

#define	NULLBYTE		0x100
#define	BLOCKSIZE	128
#define	FNAMESIZE	128
#define	MAX_FILE		256
#define	MAX_LEN		FNAMESIZE

// Transfer commands

#define	FNAMECMD		'F'
#define	BLOCKCMD		'B'
#define	READYCMD		'R'
#define	CLOSECMD		'C'
#define	DIRCMD		'D'
#define	GETCMD		'G'

// User commands

#define	ABORT			ESC
#define	LOCDIR		'L'
#define	RMTDIR		'R'
#define	XMTFILE		'T'

struct file_entry {
	char	file[FNAMESIZE];
};

char	name[FNAMESIZE];

// Prototypes

void	do_help(void);
int	kbscan(void);
int	charin(void);
bool	init_com(int baud);
int	getcomm(void);
void	putcomm(char c);
void	delay(int dly);

int	ismatch(char *pattern, char *string);
int	findfirst(char *ftemplate, struct dirent *dp);
int	findnext(struct dirent *dp);

void	sendack(void);
void	sendnak(void);
int	readfilename(void);
int	readfile(void);
int	readblock(void);
void	send_requested_file(void);
void	transmitfile(void);
int	sendfilename(char *fn);
int	sendfile(void);
int	sendblock(void);
int	getack(void);
void	add_to_filelist(char *fname);
void	showfilelist(void);
int	send_directory(void);
int	remote_directory(void);
int	local_directory(void);
int	get_local_file_list(char *template);
int	get_remote_file(void);

// Globals

struct termios	tios;

int	serialPort;
long	current_baud;
int	blocknum;
char	filename[FNAMESIZE];
byte	databfr[BLOCKSIZE];

int	filecount;
struct file_entry file_list[MAX_FILE];

char	fnames[MAX_FILE][FNAMESIZE];

DIR				*dirp;
struct dirent	finfo;
struct stat		sbuf;

FILE	*fp = NULL;

int main(void)
{
	int			i, k, done;

	if (!init_com(B9600))					// 9600 baud, 8 bits, no parity
	{
		printf("Error opening serial port\n");
		return -1;
	}

	tcgetattr(0, &tios);
	tios.c_lflag &= (~(ICANON | ECHO));
	tcsetattr(0, TCSANOW, &tios);

	current_baud = 9600;
	done = 0;

	printf("Waiting for file command...\n");

	while (!done)
	{
		k = kbscan();

		if (k)
		{
			if (k == ESC)
				done = 1;
			else
			{
				k = toupper(k);
				switch (k)
				{
					case '?':
						do_help();
						break;

					case LOCDIR:
						local_directory();
						break;

					case RMTDIR:
						remote_directory();
						break;

					case XMTFILE:
						transmitfile();
						break;

					case GETCMD:
						get_remote_file();
						break;
				}

				printf("\nWaiting for file command...\n");
			}
		}

		i = getcomm();

		if (i)
		{
			switch (i)
			{
				case FNAMECMD:
					i = readfilename();	// get file name from remote

					if (i > 0)
						readfile();			// receive file from remote

					break;

				case DIRCMD:
					send_directory();
					printf("\nWaiting for file command...\n");
					break;

				case GETCMD:
					send_requested_file();
					printf("\nWaiting for file command...\n");
					break;
			}
		}
	}

	if (fp)
	{
		fflush(fp);
		fclose(fp);
	}

	CloseDevice(serialPort);
	tios.c_lflag |= (ICANON | ECHO);
	tcsetattr(0, TCSANOW, &tios);

	return 0;
}

void do_help(void)
{
	printf("\n\n");
	printf("baud rate = %5ld\n", current_baud);
	printf("\nValid commands:\n");
	printf("<L> Local directory\n");
	printf("<R> Remote directory\n");
	printf("<T> Transmit file(s) to remote\n");
	printf("<G> Get file(s) from remote\n");
	printf("<ESC> Exit program\n\n");
}

int charin(void)
{
	int	i;

	i = 0;

	while (!i)
		i = kbscan();

	return(i);
}

int kbscan(void)
{
	int		ch;
	fd_set	readSet;
	struct timeval	timeVal;

	ch = 0;
	FD_ZERO(&readSet);		// clear the set
	FD_SET(0, &readSet);		// add this descriptor to the set
	timeVal.tv_sec = 0;
	timeVal.tv_usec = 0;

	if (select(FD_SETSIZE, &readSet, NULL, NULL, &timeVal) == 1)
	{
		ch = getc(stdin);

		if (ch == -1)
			ch = 0;
	}

	return ch;
}

// Initialize comm ports
// Return TRUE of open is successful

bool init_com(int baud)
{
	int	timeout;
	bool	cts, dcd;

	if (OpenDevice("/dev/ttyS1", &serialPort))
	{
		if (ConfigureDevice(serialPort, baud, 8, 1, 0, FALSE))
		{
			if (ConfigureFlowControl(serialPort, FALSE))
			{
				timeout = 100;

				do
				{
					GetDeviceStatus(serialPort, &cts, &dcd);

					if (cts)
						break;

					usleep(1000);
				}
				while (timeout--);

				if (!cts)
				{
					printf("No CTS\n");
					return FALSE;
				}

				ConfigureFlowControl(serialPort, TRUE);
				FlushBytes(serialPort);
				return TRUE;
			}
			else
				printf("ConfigureFlowControl failed\n");
		}
		else
			printf("ConfigureDevice failed\n");
	}
	else
		printf("OpenDevice failed\n");

	return FALSE;
}

/*	Get next character from receive buffer.
	Returns 0 if no character available else returns
	character (0x100 if received character = 0).
	If comm error occurs, bit 7 will be set.
*/

int getcomm(void)
{
	int	data;

	if (ReadBytes(serialPort, (byte *) &data, 1, 0))
	{
		data &= 0xff;

		if (!data)
			data |= 0x100;

		return data;
	}

	return 0;
}

void putcomm(char c)
{
	WriteBytes(serialPort, &c, 1);
}

int getack(void)
{
	char	c;

	c = 0;

	while (!c)
	{
		if (kbscan() == ESC)
			return 0;

		c = getcomm();
	}

	if (c == ACK)
		return 1;

	return 0;
}

void sendack(void)
{
	putcomm(ACK);
}

void sendnak(void)
{
	putcomm(NAK);
}

void send_requested_file(void)
{
	int	i, chr;

	i = readfilename();

	if (i > 0)
	{
		dirp = opendir(".");

		i = findfirst(filename, (struct dirent *) &finfo);

		if (!i)
		{
			printf("\nRemote get: %s\n", filename);
			fp = fopen((char *) &finfo.d_name, "rb");

			while (fp)
			{
				while (1)
				{
					delay(100);
					putcomm(READYCMD);
					delay(10);
					chr = getcomm();

					if (chr == ACK)
						break;

					if (kbscan() == ESC)
					{
						printf("\nuser abort\n");
						fclose(fp);
						fp = NULL;
						return;
					}
				}

				printf("\nSending file %s\n", (char *) &finfo.d_name);

				if (sendfilename((char *) &finfo.d_name))
				{
					if (sendfile())
						printf("\nError sending file\n");
				}
				else
					printf("\nError sending filename\n");

				fclose(fp);
				fp = NULL;

				i = findnext((struct dirent *) &finfo);

				if (i == -1)
					break;

				fp = fopen((char *) &finfo.d_name, "rb");
			}

			printf("\nDone sending files\n");
		}
		else
		{
			printf("\nRemote get: No files match %s\n", filename);
		}

		closedir(dirp);
	}
	else
		sendnak();
}

// Get file name from remote,
// return byte count or -1 if error.

int readfilename(void)
{
	int	i, chr, retval;

	for (i=0; i<FNAMESIZE; i++)
		filename[i] = '\0';

	blocknum = 0;
	retval = 0;
	chr = 1;
	i = 0;

	while (chr != NULLBYTE)
	{
		if (kbscan() == ESC)
		{
			retval = -1;
			break;
		}

		chr = getcomm();

		if (chr != NULLBYTE)
			chr &= 0x7f;

		if (chr == '/')
			chr = '_';

		if (isprint(chr))
			filename[i++] = (char) chr;
	}

	filename[i] = '\0';

	if (i > 0 && !retval)
		retval = i;
	else if (retval == -1)
	{
		printf("\nUser abort\n");
		sendnak();
	}
	else
	{
		printf("\nNo filename received\n");
		sendnak();
	}

	return retval;
}

//
// Receive file from remote
//

int readfile(void)
{
	int	i, bn, retval;

	blocknum = 0;
	retval = 0;
	printf("File: %s\n", filename);
	fp = fopen(filename, "wb");

	if (!fp)
	{
		printf("\nCan't open file!\n");
		sendnak();
		retval = 1;
	}
	else
	{
		sendack();

		while (1)
		{
			if (kbscan() == ESC)
			{
				retval = -1;
				break;
			}

			i = getcomm();

			if (i == BLOCKCMD)
			{
				while (1)				// get block number
				{
					i = getcomm();

					if (i)						// first digit
					{
						bn = i & 0xff;
						bn <<= 8;

						while (1)				// second digit
						{
							i = getcomm();

							if (i)
							{
								bn |= (i & 0xff);

								if (bn != blocknum)
									retval = 1;

								break;
							}
							else if (kbscan() == ESC)
							{
								retval = -1;
								break;
							}
						}

						if (!retval)
						{
							sendack();

							if (readblock())
							{
								sendnak();
								retval = 1;
								break;
							}
							else
							{
								blocknum++;
								sendack();
							}
						}
						else
							sendnak();
						break;
					}
					else if (kbscan() == ESC)
					{
						retval = -1;
						break;
					}
				}
			}
			else if (i == CLOSECMD)
			{
				fflush(fp);
				fclose(fp);
				fp = NULL;
				printf("\nTransfer complete\n");
				sendack();
				break;
			}
		}
	}

	return retval;
}

int readblock(void)
{
	int	i, chr, retval;
	char	text[16];

	printf("block: ");
	sprintf(text, "%d", blocknum);
	printf("%s", text);
	printf("\r");
	fflush(stdout);
	retval = 0;
	i = 0;

	while (i < BLOCKSIZE)
	{
		chr = getcomm();

		if (chr)
			databfr[i++] = chr & 0xff;

		if (kbscan() == ESC)
		{
			retval = -1;
			break;
		}
	}

	if (i)
		fwrite(databfr, 1, i, fp);

	return retval;
}

int sendfilename(char *fn)
{
	blocknum = 0;

	putcomm(FNAMECMD);

	while (*fn)
		putcomm(*fn++);

	putcomm(0);
	return getack();
}

int sendfile(void)
{
	int	i, retval;

	blocknum = 0;
	retval = 0;

	while (!feof(fp))
	{
		i = fread(databfr, 1, BLOCKSIZE, fp);

		if (i)
		{
			while (i < BLOCKSIZE)
				databfr[i++] = CTRLZ;
		}
		else
			break;

		if (sendblock())
		{
			retval = 1;
			break;
		}

		blocknum++;
	}

	putcomm(CLOSECMD);
	return retval;
}

int sendblock(void)
{
	int	i, retval;
	char	tempbfr[64];

	sprintf(tempbfr, "\rblock: %d", blocknum);
	printf("%s", tempbfr);
	fflush(stdout);
	retval = 0;
	putcomm(BLOCKCMD);

	while (getcomm() != READYCMD)	// wait for 'ready' response
	{
		if (kbscan() == ESC)
			return -1;
	}

	putcomm(blocknum >> 8);
	putcomm(blocknum);

	while (getcomm() != READYCMD)	// wait for 'ready' response
	{
		if (kbscan() == ESC)
			return -1;
	}

	i = 0;

	while (i < BLOCKSIZE)
	{
		putcomm(databfr[i++]);

		if (kbscan() == ESC)
		{
			retval = -1;
			break;
		}
	}

	if (!getack())
		retval = 1;

	return retval;
}

int inputfilename(void)
{
	int	i, chr, retval;

	printf("\nEnter filename: ");
	fflush(stdout);
	retval = 0;
	i = 0;

	while (1)
	{
		chr = charin();

		if (chr == ESC)
		{
			i = 0;
			break;
		}
		else if (isgraph(chr))
		{
			filename[i++] = (char) chr;
			printf("%c", chr);
		}
		else if ((chr == BACKSP || chr == X11BACKSP) && i)
		{
			printf("\b \b");
			--i;
		}
		else if (chr == '\n')
		{
			filename[i] = '\0';
			break;
		}
		fflush(stdout);
	}

	if (i)
		retval = 1;

	return retval;
}

void transmitfile(void)
{
	int	i, chr;

	printf("\nTransmit file(s) to remote.");
	fflush(stdout);

	if (inputfilename())
	{
		dirp = opendir(".");			// get pointer to directory stream

		i = findfirst(filename, (struct dirent *) &finfo);

		if (!i)
		{
			fp = fopen((char *) &finfo.d_name, "rb");

			while (fp)
			{
				while (1)
				{
					delay(100);
					putcomm(READYCMD);
					delay(10);
					chr = getcomm();

					if (chr == ACK)
						break;

					if (kbscan() == ESC)
					{
						printf("\nuser abort\n");
						fclose(fp);
						fp = NULL;
						return;
					}
				}

				printf("\nSending file %s\n", (char *) &finfo.d_name);

				if (sendfilename((char *) &finfo.d_name))
				{
					if (sendfile())
						printf("\nError sending file\n");
				}
				else
					printf("\nError sending filename\n");

				fclose(fp);
				fp = NULL;
				i = findnext((struct dirent *) &finfo);

				if (i == -1)
					break;

				fp = fopen((char *) &finfo.d_name, "rb");
			}
			printf("\nDone sending files\n");
		}
		else
		{
			printf("\nTransmit file: No files match %s\n", filename);
		}

		closedir(dirp);
	}
}

void add_to_filelist(char *fname)
{
	char	*dest;

	dest = (char *) &file_list[filecount];
	strcpy(dest, fname);
	filecount++;
}

void showfilelist(void)
{
	int	i, j, max, len;
	char	*fptr;

	printf("\n");

	for (i=0, max=0; i<filecount; i++)
	{
		fptr = (char *) &file_list[i];

		if (*fptr)
		{
			printf("%s", fptr);
			max++;

			if (max == 5)
			{
				max = 0;
				printf("\n");
			}
			else
			{
				len = strlen(fptr);

				for (j=len; j<16; j++)
					printf(" ");
			}
		}
		else
			break;
	}

	printf("\n");
}

int send_directory(void)
{
	int	i;
	char	*fptr;

	i = readfilename();

	if (i > 0)
	{
		printf("\nSend directory %s\n", filename);
		i = get_local_file_list(filename);

		if (i)
		{
			for (i=0; i<filecount; i++)
			{
				fptr = (char *) &file_list[i];

				if (*fptr)
				{
					if (!sendfilename(fptr))
					{
						printf("\nError sending file name\n");
						break;
					}
				}
				else
					break;
			}
			putcomm(CLOSECMD);
		}
		else
			putcomm(CLOSECMD);
	}
	else
	{
		printf("\nSend directory read file name failed\n");
		putcomm(NAK);
	}
	return 0;
}

int remote_directory(void)
{
	int	i, chr, retval, done;

	retval = 0;
	filecount = 0;
	printf("\nRemote directory");

	if (inputfilename())
	{
		done = FALSE;
		filecount = 0;
		putcomm(DIRCMD);
		i = 0;

		while (filename[i])			// send file template
		{
			putcomm(filename[i]);
			i++;
		}

		putcomm('\0');

		while (!done)
		{
			i = 0;
			chr = 0;

			while (!chr)
			{
				if (kbscan() == ESC)
				{
					retval = -1;
					break;
				}

				chr = getcomm();
			}

			if (chr == NAK)
			{
				done = TRUE;
				retval = 0;
				break;
			}
			else if (chr == FNAMECMD)
			{
				while (chr != NULLBYTE)
				{
					if (kbscan() == ESC)
					{
						retval = -1;
						break;
					}
					chr = getcomm();

					if (chr != NULLBYTE)
						chr &= 0x7f;

					if (chr == '/')
						chr = '_';

					if (isprint(chr))
						filename[i++] = (char) chr;
				}

				filename[i] = '\0';
				add_to_filelist(filename);
				putcomm(ACK);
			}
			else if (chr == CLOSECMD)
			{
				putcomm(ACK);
				retval = filecount;
				break;
			}
		}

		if (filecount)
			showfilelist();
		else
			printf("\nNo files\n");
	}

	return retval;
}

int get_remote_file(void)
{
	int	i, retval;

	printf("\nGet file(s) from remote.");

	retval = 0;

	if (inputfilename())
	{
		putcomm(GETCMD);
		i = 0;

		while (filename[i])			// send file template
		{
			putcomm(filename[i]);
			i++;
		}

		putcomm('\0');
		retval = getack();
	}

	return retval;
}

int get_local_file_list(char *template)
{
	int	i;

	dirp = opendir(".");			// get pointer to directory stream

	filecount = 0;
	i = findfirst(template, (struct dirent *) &finfo);

	if (!i)
	{
		add_to_filelist((char *) &finfo.d_name);
		i = 0;

		while (i != -1)
		{
			i = findnext((struct dirent *) &finfo);

			if (i != -1)
				add_to_filelist((char *) &finfo.d_name);
		}
	}

	closedir(dirp);		// done with this directory

	return filecount;
}

int local_directory(void)
{
	int				i;

	printf("\nLocal directory");

	if (inputfilename())
	{
		i = get_local_file_list(filename);

		if (i)
			showfilelist();
		else
			printf("\nNo local files\n");
	}

	return 0;
}

//
// Case insensitive match of string against pattern.
//
// Wildcard rules:
//
// 1) '?' in pattern matches any single character.
//
// 2) '*' in pattern matches any sequence of characters,
//    or none at all, until string character matches
//    the following character in pattern (if any).
//

int ismatch(char *pattern, char *string)
{
	int	i, j, wc_index, ret;
	char	next;
	char	str[MAX_LEN];
	char	pat[MAX_LEN];

	ret = FALSE;
	wc_index = -1;

	i = 0;
	while (pattern[i])					// convert pattern to upper case
	{
		pat[i] = toupper(pattern[i]);
		i++;
	}
	pat[i] = '\0';

	i = 0;
	while (string[i])						// convert string to upper case
	{
		str[i] = toupper(string[i]);
		i++;
	}
	str[i] = '\0';

	i = j = 0;
	while (1)
	{
		if (!pat[i] && !str[j])			// both empty, found a match
		{
			ret = TRUE;
			break;
		}

		if (!pat[i] && str[j])			// pattern empty, but string isn't
			break;							// no match

		if (!str[j])						// string is empty
		{
			next = pat[i + 1];			// check next in pattern
			if (!next && (pat[i] == '?' || pat[i] == '*'))
				ret = TRUE;					// match if next is wildcard
			break;
		}

		if (pat[i] == '?')				// match any single character
		{
			i++;
			j++;
		}
		else if (pat[i] == '*')			// match any sequence of chars
		{
			wc_index = i;					// save current index of wildcard
			next = pat[i + 1];			// get next pattern character
			while (str[j] && str[j] != next)
				j++;							// go through string til next match
			i++;								// then increment pattern index
		}
		else									// match next character
		{
			if (pat[i] != str[j])		// no match
			{
				if (wc_index >= 0)		// if we had a wildcard,
					i = wc_index - 1;		// back up to it
				else							// else match failed
					break;
			}
			i++;								// next pattern character
			j++;								// next string character
		}
	}

	return ret;
}

// Find first file that matches template

int findfirst(char *ftemplate, struct dirent *dp)
{
	struct dirent	*localdp;

	strcpy(name, ftemplate);

	while ((localdp = readdir(dirp)) != NULL)
	{
		lstat(localdp->d_name, &sbuf);
		if ((sbuf.st_mode & S_IFMT) != S_IFDIR)
		{
			if (ismatch(name, localdp->d_name))
			{
				memcpy(dp, localdp, sizeof(struct dirent));
				return 0;
			}
		}
	}

	return -1;
}

// Find next file matching original template

int findnext(struct dirent *dp)
{
	struct dirent	*localdp;

	while ((localdp = readdir(dirp)) != NULL)
	{
		lstat(localdp->d_name, &sbuf);
		if ((sbuf.st_mode & S_IFMT) != S_IFDIR)
		{
			if (ismatch(name, localdp->d_name))
			{
				memcpy(dp, localdp, sizeof(struct dirent));
				return 0;
			}
		}
	}

	return -1;
}

void delay(int dly)
{
	usleep(dly * 1000);	// convert usec to msec and wait
}

// end of file


